@@ -4,7 +4,7 @@
 # for information about the DNS protocol. This should insulate Tcl scripts
 # from problems with using the system library resolver for slow name servers.
 #
-# This implementation uses TCP only for DNS queries. The protocol reccommends
+# This implementation uses TCP only for DNS queries. The protocol recommends
 # that UDP be used in these cases but Tcl does not include UDP sockets by
 # default. The package should be simple to extend to use a TclUDP extension
 # in the future.
@@ -15,6 +15,7 @@
 # Support added for RFC1886 - DNS Extensions to support IP version 6
 # Support added for RFC2782 - DNS RR for specifying the location of services
 # Support added for RFC1995 - Incremental Zone Transfer in DNS
+# Support added for RFC7858 - DNS over Transport Layer Security
 #
 # TODO:
 #  - When using tcp we should make better use of the open connection and
@@ -28,8 +29,6 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 # -------------------------------------------------------------------------
-#
-# $Id: dns.tcl,v 1.36 2008/11/22 12:28:54 mic42 Exp $
 
 package require Tcl 8.2;                # tcl minimum version
 package require logger;                 # tcllib 1.3
@@ -38,9 +37,6 @@ package require uri::urn;               # tcllib 1.2
 package require ip;                     # tcllib 1.7
 
 namespace eval ::dns {
-    variable version 1.3.4
-    variable rcsid {$Id: dns.tcl,v 1.36 2008/11/22 12:28:54 mic42 Exp $}
-
     namespace export configure resolve name address cname \
         status reset wait cleanup errorcode
 
@@ -53,6 +49,12 @@ namespace eval ::dns {
             search     {}
             nameserver {localhost}
             loglevel   warn
+            usetls     0
+            cafile     ""
+            cadir      ""
+        }
+        if {[file exists /etc/ssl/certs/ca-certificates.crt]} {
+            set options(cafile) /etc/ssl/certs/ca-certificates.crt
         }
         variable log [logger::init dns]
         ${log}::setlevel $options(loglevel)
@@ -69,12 +71,13 @@ namespace eval ::dns {
     }
 
     variable types
-    array set types { 
-        A 1  NS 2  MD 3  MF 4  CNAME 5  SOA 6  MB 7  MG 8  MR 9 
+    array set types {
+        A 1  NS 2  MD 3  MF 4  CNAME 5  SOA 6  MB 7  MG 8  MR 9
         NULL 10  WKS 11  PTR 12  HINFO 13  MINFO 14  MX 15  TXT 16
-        SPF 16 AAAA 28 SRV 33 IXFR 251 AXFR 252  MAILB 253  MAILA 254
+        SPF 16 AAAA 28 SRV 33 NAPTR 35 IXFR 251 AXFR 252 MAILB 253
+        MAILA 254
         ANY 255 * 255
-    } 
+    }
 
     variable classes
     array set classes { IN 1  CS 2  CH  3  HS 4  * 255}
@@ -107,25 +110,25 @@ proc ::dns::configure {args} {
     if {[llength $args] == 1} {
         set cget 1
     }
-   
+
     while {[string match -* [lindex $args 0]]} {
         switch -glob -- [lindex $args 0] {
             -n* -
             -ser* {
                 if {$cget} {
-                    return $options(nameserver) 
+                    return $options(nameserver)
                 } else {
-                    set options(nameserver) [Pop args 1] 
+                    set options(nameserver) [Pop args 1]
                 }
             }
-            -po*  { 
+            -po*  {
                 if {$cget} {
                     return $options(port)
                 } else {
-                    set options(port) [Pop args 1] 
+                    set options(port) [Pop args 1]
                 }
             }
-            -ti*  { 
+            -ti*  {
                 if {$cget} {
                     return $options(timeout)
                 } else {
@@ -142,14 +145,14 @@ proc ::dns::configure {args} {
                         return -code error "invalid protocol \"$proto\":\
                             protocol must be either \"udp\" or \"tcp\""
                     }
-                    set options(protocol) $proto 
+                    set options(protocol) $proto
                 }
             }
-            -sea* { 
+            -sea* {
                 if {$cget} {
                     return $options(search)
                 } else {
-                    set options(search) [Pop args 1] 
+                    set options(search) [Pop args 1]
                 }
             }
             -log* {
@@ -160,6 +163,20 @@ proc ::dns::configure {args} {
                     ${log}::setlevel $options(loglevel)
                 }
             }
+            -cafile {
+                if {$cget} {
+                    return $options(cafile)
+                } else {
+                    set options(cafile) [Pop args 1]
+                }
+            }
+            -cadir {
+                if {$cget} {
+                    return $options(cadir)
+                } else {
+                    set options(cadir) [Pop args 1]
+                }
+            }
             --    { Pop args ; break }
             default {
                 set opts [join [lsort [array names options]] ", -"]
@@ -207,6 +224,9 @@ proc ::dns::resolve {query args} {
     set state(-port)       $options(port);      # default namerservers port
     set state(-search)     $options(search);    # domain search list
     set state(-protocol)   $options(protocol);  # which protocol udp/tcp
+    set state(-usetls)     $options(usetls);    # use RFC7858 privacy
+    set state(-cafile)     $options(cafile);    # certificate authority file
+    set state(-cadir)      $options(cadir);     # certificate authority dir
 
     # Handle DNS URL's
     if {[string match "dns:*" $query]} {
@@ -214,7 +234,7 @@ proc ::dns::resolve {query args} {
         foreach {opt value} [uri::split $query] {
             if {$value != {} && [info exists state(-$opt)]} {
                 set state(-$opt) $value
-            }   
+            }
         }
         set state(query) $URI(query)
         ${log}::debug "parsed query: $query"
@@ -225,6 +245,9 @@ proc ::dns::resolve {query args} {
             -n* - ns -
             -ser* { set state(-nameserver) [Pop args 1] }
             -po*  { set state(-port) [Pop args 1] }
+            -usetls { set state(-usetls) [Pop args 1] }
+            -cafile { set state(-cafile) [Pop args 1] }
+            -cadir { set state(-cadir) [Pop args 1] }
             -ti*  { set state(-timeout) [Pop args 1] }
             -co*  { set state(-command) [Pop args 1] }
             -cl*  { set state(-class) [Pop args 1] }
@@ -248,6 +271,14 @@ proc ::dns::resolve {query args} {
         return -code error "no nameserver specified"
     }
 
+    if {$state(-usetls)} {
+        package require tls
+        set state(-protocol) "tcp"
+        if {$state(-port) == $options(port)} {
+            set state(-port) 853
+        }
+    }
+
     if {$state(-protocol) == "udp"} {
         if {[llength [package provide ceptcl]] == 0 \
                 && [llength [package provide udp]] == 0} {
@@ -255,26 +286,32 @@ proc ::dns::resolve {query args} {
                 get ceptcl or tcludp"
         }
     }
-    
-    # Check for reverse lookups
+
+    # Check for reverse lookups. IPv4 first, then IPv6.
     if {[regexp {^(?:\d{0,3}\.){3}\d{0,3}$} $state(query)]} {
         set addr [lreverse [split $state(query) .]]
         lappend addr in-addr arpa
         set state(query) [join $addr .]
         set state(-type) PTR
+    } elseif {[string match {*:*} $state(query)]} {
+        set addr [ip::normalize $state(query)]
+        set addr [split [string reverse $addr] :]
+        set addr [join [split [join $addr ""] {}] .]
+        lappend addr ip6 arpa
+        set state(query) [join $addr .]
+        set state(-type) PTR
     }
 
     BuildMessage $token
-    
+
     if {$state(-protocol) == "tcp"} {
         TcpTransmit $token
-        if {$state(-command) == {}} {
-            wait $token
-        }
     } else {
         UdpTransmit $token
     }
-    
+    if {$state(-command) == {}} {
+        wait $token
+    }
     return $token
 }
 
@@ -460,7 +497,7 @@ proc ::dns::dump {args} {
     # FRINK: nocheck
     variable $token
     upvar 0 $token state
-    
+
     set result {}
     switch -glob -- $type {
         -qu*    -
@@ -543,7 +580,7 @@ proc ::dns::BuildMessage {token} {
             append state(request) $qsection $nsdata
         }
         1 {
-            # IQUERY            
+            # IQUERY
             set state(request) [binary format SSSSSS $state(id) \
                 [expr {($state(opcode) << 11) | ($state(-recurse) << 8)}] \
                 0 $qdcount 0 0 0]
@@ -603,9 +640,9 @@ proc ::dns::PackRecord {args} {
 
     switch -exact -- $rr(type) {
         CNAME - MB - MD - MF - MG - MR - NS - PTR {
-            set rr(rdata) [PackName $rr(rdata)] 
+            set rr(rdata) [PackName $rr(rdata)]
         }
-        HINFO { 
+        HINFO {
             array set r {CPU {} OS {}}
             array set r $rr(rdata)
             set rr(rdata) [PackString $r(CPU)]
@@ -630,7 +667,7 @@ proc ::dns::PackRecord {args} {
                 set s [string range $str $n [incr n 253]]
                 append rr(rdata) [PackString $s]
             }
-        }          
+        }
         NULL {}
         SOA {
             array set r {MNAME {} RNAME {}
@@ -689,15 +726,51 @@ proc ::dns::TcpConnected {token s} {
 	return
     }
 
+    if {$state(-usetls)} {
+        tls::import $s -server false -request 1 \
+            -cadir $state(-cadir) \
+            -cafile $state(-cafile) \
+            -ssl2 false -ssl3 false -tls1 true \
+            -command [list ::dns::TlsCallback $token]
+        if {[catch {tls::handshake $s} err]} {
+            Finish $token $err
+            return
+        }
+    }
+
     fconfigure $s -blocking 0 -translation binary -buffering none
 
     # For TCP the message must be prefixed with a 16bit length field.
     set req [binary format S [string length $state(request)]]
     append req $state(request)
 
-    puts -nonewline $s $req
-
     fileevent $s readable [list [namespace current]::TcpEvent $token]
+    puts -nonewline $s $req
+}
+
+proc ::dns::TlsCallback {token cmd channel args} {
+    variable log
+    variable $token
+    upvar 0 $token state
+    switch -exact -- $cmd {
+        info {
+            foreach {major minor message} $args break
+            ${log}::debug "TLS: $major/$minor $message"
+        }
+        verify {
+            foreach {depth cert status error} $args break
+            lappend state(certChain) \
+                [list depth $depth status $status error $error cert $cert]
+            return $status
+        }
+        error {
+            return -code error "tls error: $args"
+        }
+        default {
+            return -code error "unexpected message type \"$cmd\" in TLS callback"
+        }
+    }
+    return 1
 }
 
 # -------------------------------------------------------------------------
@@ -721,7 +794,7 @@ proc ::dns::UdpTransmit {token} {
                                    $token timeout\
                                   "operation timed out"]]
     }
-    
+
     if {[llength [package provide ceptcl]] > 0} {
         # using ceptcl
         set state(sock) [cep -type datagram $state(-nameserver) $state(-port)]
@@ -734,9 +807,9 @@ proc ::dns::UdpTransmit {token} {
     fconfigure $state(sock) -translation binary -buffering none
     set state(status) connect
     puts -nonewline $state(sock) $state(request)
-    
+
     fileevent $state(sock) readable [list [namespace current]::UdpEvent $token]
-    
+
     return $token
 }
 
@@ -758,7 +831,9 @@ proc ::dns::Finish {token {errormsg ""}} {
     catch {close $state(sock)}
     catch {after cancel $state(after)}
     if {[info exists state(-command)] && $state(-command) != {}} {
-	if {[catch {eval $state(-command) {$token}} err]} {
+	if {[catch {
+            uplevel #0 [linsert $state(-command) end $token]
+        } err]} {
 	    if {[string length $errormsg] == 0} {
 		set state(error) [list $err $errorInfo $errorCode]
 		set state(status) error
@@ -799,7 +874,7 @@ proc ::dns::Receive {token} {
     switch -- $status {
         0 {
             set state(status) ok
-            Finish $token 
+            Finish $token
         }
         1 { Finish $token "Format error - unable to interpret the query." }
         2 { Finish $token "Server failure - internal server error." }
@@ -834,14 +909,15 @@ proc ::dns::TcpEvent {token} {
         ${log}::debug "Event error: $result"
         Finish $token "error reading data: $result"
     } elseif { [string length $result] >= 0 } {
+        ${log}::debug "read [string length $result] bytes for $token"
         if {[catch {
             # Handle incomplete reads - check the size and keep reading.
             if {![info exists state(size)]} {
                 binary scan $result S state(size)
-                set result [string range $result 2 end]            
+                set result [string range $result 2 end]
             }
             append state(reply) $result
-            
+
             # check the length and flags and chop off the tcp length prefix.
             if {[string length $state(reply)] >= $state(size)} {
                 binary scan $result S id
@@ -891,14 +967,14 @@ proc ::dns::UdpEvent {token} {
     #Receive [namespace current]::$id
     Receive $token
 }
-    
+
 # -------------------------------------------------------------------------
 
 proc ::dns::Flags {token {varname {}}} {
     # FRINK: nocheck
     variable $token
     upvar 0 $token state
-    
+
     if {$varname != {}} {
         upvar $varname flags
     }
@@ -1051,7 +1127,7 @@ proc ::dns::ReadQuestion {nitems data indexvar} {
         set r {}
         lappend r name [ReadName data $index offset]
         incr index $offset
-        
+
         # Read off QTYPE and QCLASS for this query.
         set ndx $index
         incr index 3
@@ -1065,10 +1141,10 @@ proc ::dns::ReadQuestion {nitems data indexvar} {
     }
     return $result
 }
-        
+
 # -------------------------------------------------------------------------
 
-# Read an answer section from a DNS message. 
+# Read an answer section from a DNS message.
 #
 proc ::dns::ReadAnswer {nitems data indexvar {raw 0}} {
     variable types
@@ -1080,7 +1156,7 @@ proc ::dns::ReadAnswer {nitems data indexvar {raw 0}} {
         set r {}
         lappend r name [ReadName data $index offset]
         incr index $offset
-        
+
         # Read off TYPE, CLASS, TTL and RDLENGTH
         binary scan [string range $data $index end] SSIS type class ttl rdlength
 
@@ -1104,7 +1180,7 @@ proc ::dns::ReadAnswer {nitems data indexvar {raw 0}} {
                     set rdata [ip::contract [ip::ToString $rdata]]
                 }
                 NS - CNAME - PTR {
-                    set rdata [ReadName data $index off] 
+                    set rdata [ReadName data $index off]
                 }
                 MX {
                     binary scan $rdata S preference
@@ -1120,15 +1196,38 @@ proc ::dns::ReadAnswer {nitems data indexvar {raw 0}} {
                     lappend rdata port [ReadUShort data $x off]
                     incr x $off
                     lappend rdata target [ReadName data $x off]
-                    incr x $off
                 }
+		NAPTR {
+		    set x $index
+		    set rdata [list order [ReadUShort data $x off]]
+		    incr x $off
+		    lappend rdata preference [ReadUShort data $x off]
+		    incr x $off
+		    lappend rdata flags [ReadString data $x off]
+		    incr x $off
+		    lappend rdata service [ReadString data $x off]
+		    incr x $off
+		    lappend rdata regex [ReadString data $x off]
+		    incr x $off
+		    set domain {}
+		    while {$x < $index + $rdlength} {
+			lappend domain [ReadString data $x off]
+			incr x $off
+		    }
+		    lappend rdata replacement [join $domain .]
+		}
                 TXT {
-                    set rdata [ReadString data $index $rdlength]
+		    set x $index
+		    set rdata ""
+		    while {$x < $index + $rdlength} {
+			append rdata [ReadString data $x off]
+			incr x $off
+		    }
                 }
                 SOA {
                     set x $index
                     set rdata [list MNAME [ReadName data $x off]]
-                    incr x $off 
+                    incr x $off
                     lappend rdata RNAME [ReadName data $x off]
                     incr x $off
                     lappend rdata SERIAL [ReadULong data $x off]
@@ -1154,7 +1253,7 @@ proc ::dns::ReadAnswer {nitems data indexvar {raw 0}} {
 
 
 # Read a 32bit integer from a DNS packet. These are compatible with
-# the ReadName proc. Additionally - ReadULong takes measures to ensure 
+# the ReadName proc. Additionally - ReadULong takes measures to ensure
 # the unsignedness of the value obtained.
 #
 proc ::dns::ReadLong {datavar index usedvar} {
@@ -1176,8 +1275,8 @@ proc ::dns::ReadULong {datavar index usedvar} {
     if {[binary scan $data @${index}cccc b1 b2 b3 b4]} {
         set used 4
         # This gets us an unsigned value.
-        set r [expr {($b4 & 0xFF) + (($b3 & 0xFF) << 8) 
-                     + (($b2 & 0xFF) << 16) + ($b1 << 24)}] 
+        set r [expr {($b4 & 0xFF) + (($b3 & 0xFF) << 8)
+                     + (($b2 & 0xFF) << 16) + ($b1 << 24)}]
     }
     return $r
 }
@@ -1190,12 +1289,12 @@ proc ::dns::ReadUShort {datavar index usedvar} {
     if {[binary scan [string range $data $index end] cc b1 b2]} {
         set used 2
         # This gets us an unsigned value.
-        set r [expr {(($b2 & 0xff) + (($b1 & 0xff) << 8)) & 0xffff}] 
+        set r [expr {(($b2 & 0xff) + (($b1 & 0xff) << 8)) & 0xffff}]
     }
     return $r
 }
 
-# Read off the NAME or QNAME element. This reads off each label in turn, 
+# Read off the NAME or QNAME element. This reads off each label in turn,
 # dereferencing pointer labels until we have finished. The length of data
 # used is passed back using the usedvar variable.
 #
@@ -1207,13 +1306,13 @@ proc ::dns::ReadName {datavar index usedvar} {
     set r {}
     set len 1
     set max [string length $data]
-    
+
     while {$len != 0 && $index < $max} {
         # Read the label length (and preread the pointer offset)
         binary scan [string range $data $index end] cc len lenb
         set len [expr {$len & 0xFF}]
         incr index
-        
+
         if {$len != 0} {
             if {[expr {$len & 0xc0}]} {
                 binary scan [binary format cc [expr {$len & 0x3f}] [expr {$lenb & 0xff}]] S offset
@@ -1230,23 +1329,23 @@ proc ::dns::ReadName {datavar index usedvar} {
     return [join $r .]
 }
 
-proc ::dns::ReadString {datavar index length} {
+proc ::dns::ReadString {datavar index usedvar} {
     upvar $datavar data
+    upvar $usedvar used
     set startindex $index
 
     set r {}
-    set max [expr {$index + $length}]
 
-    while {$index < $max} {
-        binary scan [string range $data $index end] c len
-        set len [expr {$len & 0xFF}]
-        incr index
+    if {[binary scan [string range $data $index end] c len] == 1} {
+	set len [expr {$len & 0xFF}]
+	incr index
 
-        if {$len != 0} {
-            append r [string range $data $index [expr {$index + $len - 1}]]
-            incr index $len
-        }
+	if {$len != 0} {
+	    set r [string range $data $index [expr {$index + $len - 1}]]
+	    incr index $len
+	}
     }
+    set used [expr {$index - $startindex}]
     return $r
 }
 
@@ -1374,11 +1473,11 @@ proc ::uri::SplitDns {uri} {
         # Handle the nameserver specification
         if {[string match "//*" $uri]} {
             set uri [string range $uri 2 end]
-            array set tmp [GetHostPort uri]
+            array set tmp [GetUPHP uri]
             set parts(nameserver) $tmp(host)
             set parts(port) $tmp(port)
         }
-        
+
         # what's left is the query domain name.
         set parts(query) [string trimleft $uri /]
     }
@@ -1414,7 +1513,7 @@ proc ::uri::JoinDns {args} {
 
 catch {dns::configure -nameserver [lindex [dns::nameservers] 0]}
 
-package provide dns $dns::version
+package provide dns 1.5.0
 
 # -------------------------------------------------------------------------
 # Local Variables:
