@@ -8,6 +8,8 @@
 #include <tc.h>    /* Token cache handling */
 #include <util.h>  /* Allocation utilities */
 #include <string.h>
+#include <ctype.h>  /* is... */
+#include <stdlib.h> /* qsort */
 
 /*
  * = = == === ===== ======== ============= =====================
@@ -423,7 +425,7 @@ rde_param_query_ls (RDE_PARAM p, long int* lc, void*** lv)
 SCOPE long int
 rde_param_query_lstop (RDE_PARAM p)
 {
-    (long int) rde_stack_top (p->LS);
+    return (long int) rde_stack_top (p->LS);
 }
 
 SCOPE Tcl_HashTable*
@@ -560,13 +562,14 @@ rde_param_i_error_nonterminal (RDE_PARAM p, long int s)
      * directly call on this function.
      */
     return;
-
+#if 0
     long int pos;
     if (!p->ER) return;
     pos = 1 + (long int) rde_stack_top (p->LS);
     if (p->ER->loc != pos) return;
     error_set (p, s);
     p->ER->loc = pos;
+#endif
 }
 
 SCOPE void
